(* 
 * Copyright (c) 2001, 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: gont.ml,v 1.7 2002/02/25 11:08:48 malekith Exp $
 *
 * Module implementation: Gont
 * Description: Top level module for Gont compiler. There are two main entry
 * points here. First one, 'compile' compiles specified source file, the 
 * second 'compile_iface' compiles interface. 
 *
 * Compilation of source file involves:
 *  - initialization of Ksi code generator
 *  - opening input and output files
 *  - parsing it -- we parse one declaration at a time, and generate code
 *    for it. This generally involves sequence of:
 *      - Gontbind.bind -- binds names to symbols, reports any 
 *        undefined symbols
 *      - Gonttypes.typecheck_decl -- checks whatever declaration is 
 *        typecorrect
 *      - Gontcodegen.ksi_code -- returns list of Ksi declarations to
 *        be output (we might generate more then one Ksi declarations
 *        from given Gont declaration, hence list).
 *  - closing input and output files
 * 
 * Compilation of interface, is at the very moment simplified. It does
 * not check whatever given declaration has already been seen (we should
 * report error then), and few other things. We generally:
 *  - open input file
 *  - parse it, this results in list of declarations
 *  - pass it to Gontmodule.save function, that makes use of OCaml 
 *    serialization code.
 *  - close input file
 *)

type compilation_info = {ci_input_file: string;
                         ci_output_file: string;
                         ci_goutput_file: string;
                         ci_target_prefix: string;
                         ci_module_name: string}
open Gonttree
open Gontutil
 
type compilation_ctx = {ctx_out_ch: out_channel;
                        ctx_iface: bool;
                        ctx_bind_ctx: Gontbind.ctx;
                        ctx_codegen_ctx: Gontcodegen.ctx option;
                        ctx_lexbuf: Lexing.lexbuf;
                        ctx_module: string}
                        
let loc (f, l, c) = {loc_file = f; loc_line = l; loc_char = c}

let output_ksi ctx decls =
   List.iter (Ksi.output_decl ctx.ctx_out_ch) decls

let parse_decl ctx =
  let lexbuf = ctx.ctx_lexbuf in
  let iface = ctx.ctx_iface in
    try
      if iface then
        (Parser.iface_decl Lexer.token lexbuf) 
      else
        (Parser.decl Lexer.token lexbuf) 
    with 
        Parsing.Parse_error ->
          raise(Syntax_error (loc (Lexer.get_current_location ()),
                              "parse error"))
      | Lexer.Eof ->
          raise Eof
      | Lexer.Error(f, l, c, e) ->
          raise(Syntax_error (loc (f, l, c), e))

let compile_decl ctx decl =
  let iface = ctx.ctx_iface in
  let bind_ctx = Gontbind.bind ctx.ctx_bind_ctx decl in
  let _ = Gonttypes.typecheck_decl decl in
  if iface then
    {ctx with ctx_bind_ctx = bind_ctx}
  else
    let (codegen_ctx, code) =
        match ctx.ctx_codegen_ctx with
        | Some ctx -> Gontcodegen.ksi_code ctx decl
        | None -> assert false in
    let _ = output_ksi ctx code in
    {ctx with ctx_bind_ctx = bind_ctx; 
              ctx_codegen_ctx = Some codegen_ctx}

let rec compile_all ctx =
  try
    let ctx = compile_decl ctx (parse_decl ctx) in
    compile_all ctx
  with Eof -> ctx 
      
let compile_init_fini ctx =
  let loc = loc (Lexer.get_current_location ()) in
  let decl name body =
    FuncDecl {fn_ret_type = Void;
              fn_name = name;
              fn_parms = [];
              fn_body = {raw_stmt = Block body; stmt_loc = loc};
              fn_loc = loc;
              fn_tyvars = [];
              fn_symbol = None;
              fn_export_as = ""} in
  let ctx = 
    match (Gontbind.init_stmts ctx.ctx_bind_ctx) with
    | [] -> ctx
    | body -> compile_decl ctx (decl "init$" body) in
  let ctx = 
    match (Gontbind.fini_stmts ctx.ctx_bind_ctx) with
    | [] -> ctx
    | body -> compile_decl ctx (decl "fini$" body) in
  ctx

let do_compile iface info =
  let in_ch =
    if info.ci_input_file = "-" then 
       stdin 
     else 
       open_in info.ci_input_file in
  let out_ch = 
    if info.ci_output_file = "-" then 
       stdout
     else 
       open_out info.ci_output_file in
       
  let close_files () =
     if out_ch != stdout then close_out out_ch else ();
     if in_ch != stdin then close_in in_ch else () in

  try
    Gontutil.reset ();
    Gontmodule.reset ();
    
    let bind_ctx = Gontbind.init_bind info.ci_module_name iface in
    let codegen_ctx = 
        if iface then
           None
        else           
           Some (Gontcodegen.init_codegen info.ci_module_name
                                          info.ci_input_file) in
    let lexbuf = Lexing.from_channel in_ch in
    let ctx = {ctx_out_ch = out_ch;
               ctx_iface = iface;
               ctx_bind_ctx = bind_ctx;
               ctx_lexbuf = lexbuf;
               ctx_codegen_ctx = codegen_ctx;
               ctx_module = info.ci_module_name} in
    
    Lexer.set_file_name info.ci_input_file;

    if not iface then begin
      Ksi.read_target_conf info.ci_target_prefix;
      Ksi.output_header out_ch;
      match codegen_ctx with
      | Some codegen_ctx -> 
        output_ksi ctx (Gontcodegen.ksi_runtime_proto codegen_ctx)
      | None -> assert false
    end;

    let ctx = 
       if iface then
          if info.ci_module_name <> "Std" then
               compile_decl ctx (TopLevelOpen "Std")
          else
               ctx
         else
           compile_decl ctx (TopLevelOpen "Std") in
    
    let ctx = compile_all ctx in

    if iface then
        Gontbind.finalize_module ctx.ctx_out_ch ctx.ctx_bind_ctx
    else begin
      begin
        let ctx = compile_init_fini ctx in
        match ctx.ctx_codegen_ctx with
        | Some ctx -> 
            Gontcodegen.check_module ctx;
            let mi = Gontcodegen.get_module_info ctx in
            let go_ch = open_out info.ci_goutput_file in
            Gontmodule.save_module_info go_ch mi;
            close_out go_ch
        | None -> assert false
      end;
      Ksi.output_trailer out_ch
    end;
    
    close_files ()
  with x -> 
    close_files ();
    raise x

let do_compile_no_exn iface info =
  try
    let _ = do_compile iface info in
    0
  with
    | Syntax_error(loc, msg) ->
        err_msg (loc, msg);
        1
    | Error(msg) ->
        err_msg ({loc_file=info.ci_input_file; loc_line=0; loc_char=0}, msg);
        1
    | x ->
        prerr_endline ("Unhandled exception: " ^ (Printexc.to_string x));
        1

let compile = do_compile_no_exn false
let compile_iface = do_compile_no_exn true
;;

let dump_module m =
  print_endline (xf "// module dump of %s" m);
  let (g, t) = Gontmodule.globals m in
  List.iter (fun t -> 
                print_endline (xf "type %s = %s;" t.td_name
                                  (Gonttypes.string_of_type t.td_type)))
           t;
  List.iter (fun sym -> 
                print_endline (xf "%s %s;" (Gonttypes.string_of_type
                               sym.sym_type) sym.sym_name))
           g;
  0
;;
